#!/bin/bash

set -e

helper_basename="$(basename "$0")"
const_pkg='pkg'
const_run_skill="$const_pkg/run/sbcl/skill"
const_xml='testing-result.xml'
const_error='*** ERROR ***'
const_work_dir='testing-work'

opt_clone_base=
opt_requires=
opt_skill=
opt_triggers=
opt_nice='<auto>'
opt_report_format='xml'
opt_assume_ready='no'

this_pkg=
pkgs_ready=
preloaded_pkgs=
mk_work_dir=1
pkg_to_this_pkg='../..'

function usage {
    message="$*"

    cat <<EOF
Usage:
  $helper_basename <command> [<args>]

Commands:
  test                    Check out dependencies in work dir and run tests;
  test-in <pkg>           Test local package in directory <pkg>;
  with-rhino <cmd>        Execute command <cmd> within a fresh Rhino impl;

Common options:
  --nice <cmd>            Prefix interpreter invocations with <cmd>;
  --clone-base <dir>      Sets directory for cloning packages from to <dir>;
  --require <pkg>         Check out and load package <pkg> before this one;

Options for test, test-in:
  --skill <impl>          Use SKILL implementation <impl> (default: fresh);
  --report-format <fmt>   Generate report in format <fmt>;

Options for test-in:
  --assume-ready <yes|no> Assume required packages are ready;

Options for jenkins-job:
  --triggers <child>      Building this project triggers project <child>.

EOF

    if test -n "$message"; then
        echo "$message"
        echo
    fi
}

function die_usage {
    local status="$1"; shift
    usage "$const_error $*"
    exit "$status"
}

function die {
    local status="$1"; shift
    echo
    echo "$const_error $*"
    echo
    exit "$status"
}

function read_pkg_requires {
    local head="$1"; shift

    < "$head" sed -n                            \
        -e '/;[; \t]*Code:/q'                   \
        -e '/;[; \t]*Commentary:/q'             \
        -e 's/;[; \t]*@requires//p'
}

function pkg_requires {
    local pkg="$1"; shift
    local pkg_dir="$1"; shift
    local head="$pkg_dir/$pkg.il"

    if ! test -f "$head"; then
        head="$pkg_dir/$pkg.ils"
    fi

    local requires="$(read_pkg_requires "$head")"
    if test -n "$requires"; then
        echo "$requires"
    else
        echo '/'
    fi
}

function ensure_pkgs_ready {
    if test "$opt_assume_ready" = 'yes'; then
        return
    fi

    local work_dir="$1"; shift
    local pkgs="$1"; shift
    local reqs=''

    for pkg in ${pkgs//,/ }; do
        case "$pkgs_ready" in
            */$pkg/*)
                echo "Package "$pkg" already checked-out; ignoring."
                ;;
            *)
                local pkg_dir="$work_dir/$const_pkg/$pkg"
                git clone "$opt_clone_base/$pkg" "$pkg_dir"
                pkgs_ready="$pkgs_ready$pkg/"
                local prs="$(pkg_requires "$pkg" "$pkg_dir/")"
                if test "$prs" != '/'; then
                    reqs="$reqs,$prs"
                fi
                ;;
        esac
    done

    if test -n "${reqs//,/}"; then
        ensure_pkgs_ready "$work_dir" "$reqs"
    fi
}

function pkg_set_globals {
    local clone_base=$(git config remote.origin.url)
    local this_suffix="$(basename "$clone_base")"
    local origin_base="$(dirname "$clone_base")"

    # Set global package name
    if test -z "$this_pkg"; then
        this_pkg="${this_suffix/.git/}"
    fi

    # Set clone base if necessary.
    if test -z "$opt_clone_base"; then
        opt_clone_base="$origin_base"
    fi

    if test "$opt_nice" == '<auto>'; then
        local nice=' '
        if which nice > /dev/null; then
            nice=' nice'
        fi
        if which ionice > /dev/null; then
            nice="${nice} ionice -c 3"
        fi
        opt_nice="${nice:1}"
    fi
}

function pkg_bootstrap {
    if test -n "$pkgs_ready"; then
        return
    fi

    local work_dir="$1"; shift
    pkg_set_globals

    if test "$mk_work_dir" == 1; then
        mk_work_dir=0
        mkdir "$work_dir/"
        mkdir "$work_dir/$const_pkg/"
    else
        mkdir -p "$work_dir/$const_pkg/"
        rm -f "$work_dir/$const_pkg/$this_pkg"
    fi

    # Create self-link.
    (cd "$work_dir/$const_pkg/" && ln -s "$pkg_to_this_pkg" "$this_pkg")
    pkgs_ready="/$this_pkg/"
    local pkg_dir="$work_dir/$const_pkg/$this_pkg"
    local prs="$(pkg_requires "$this_pkg" "$pkg_dir")"
    if test "$prs" != '/'; then
        ensure_pkgs_ready "$work_dir" "$prs"
    fi

    if test -z "$opt_skill"; then
        preloaded_pkgs='run testing'
        ensure_pkgs_ready "$work_dir" "$preloaded_pkgs"
        $opt_nice sbcl --script "$work_dir/$const_pkg/run/sbcl/cmd.lisp" build
        test -x "$work_dir/$const_run_skill"
        opt_skill="./$const_run_skill"
    else
        preloaded_pkgs='testing'
        ensure_pkgs_ready "$work_dir" "$preloaded_pkgs"
    fi
}

function parse_options {
    local cmd="$1"; shift

    while test -n "$1"; do
        case "$cmd,$1" in
            *,--nice)
                opt_nice="$2"
                shift 2
                ;;
            *,--clone-base)
                opt_clone_base="$2"
                shift 2
                ;;
            *,--require)
                opt_requires="$opt_requires,$2"
                shift 2
                ;;
            test*,--skill)
                opt_skill="$2"
                shift 2
                ;;
            test*,--report-format)
                opt_report_format="$2"
                shift 2
                ;;
            test-in,--assume-ready)
                opt_assume_ready="$2"
                shift 2
                ;;
            jenkins-job,--triggers)
                opt_triggers="$opt_triggers,$2"
                shift 2
                ;;
            *)
                die_usage 2 "Bad argument/option for $cmd: $1; tail: $*."
                ;;
        esac
    done
}

function test_in {
    local work_dir="$1"; shift
    local boot_prefixes="$1"; shift
    local test_pkg="$1"; shift
    local requires="$1"; shift

    (
        cd "$work_dir"
        test -f "$const_xml" && rm "$const_xml"
        $opt_nice "$opt_skill" <<EOF
(define XtsGlobalSuccess nil)
(progn
  (exists prefix (parseString "$boot_prefixes")
    (let ((boot (strcat prefix "/testing/testing.ils")))
      (when (isFile boot)
        (load boot))))
  (let ((base_opts (quote (nil
                           chattyp t
                           reportFilename "./$const_xml")))
        (other_opts (list (quote reportFormat) (concat "$opt_report_format")
                          (quote _preloaded) (parseString "$preloaded_pkgs"))))
    (let ((r (VedaTestingRunTestsOnPackage "$test_pkg"
                                           ?options (append base_opts
                                                            other_opts))))
      (setq XtsGlobalSuccess r))))
(unless XtsGlobalSuccess
  (printf "\\n$const_error Test failures occured.\\n")
  (exit 1))
EOF

        if test "$opt_report_format" = 'xml'; then
            test -f "$const_xml" ||                                    \
                die 3 "Missing test results; package: '$test_pkg'; "    \
                "implementation: '$opt_skill'."
        fi
    )
}

function cmd_test {
    parse_options 'test' "$@"

    local requires="$opt_requires"

    pkg_bootstrap "$const_work_dir"
    ensure_pkgs_ready "$const_work_dir" "$requires"
    test_in "$const_work_dir" "$const_pkg" "$this_pkg" "$requires"
}

function cmd_test_in {
    local work_dir="$1"; shift
    test -n "$work_dir" || \
        die_usage 2 "Command test-in takes a mandatory argument."
    local requires="$opt_requires"
    local resolved_dir="$(cd "$work_dir" && /bin/pwd)"

    this_pkg="$(basename "$resolved_dir")"

    parse_options 'test-in' "$@"

    # KLUDGE: We are changing run directories, so won't benefit from
    # any package checked-out by with-rhino.  This obviously breaks
    # down with test-in ., but we don't care for now.  A better
    # solution would be to use a specific package path.
    pkgs_ready=''

    mk_work_dir=0
    pkg_to_this_pkg='..'
    pkg_bootstrap "$work_dir"
    ensure_pkgs_ready "$work_dir" "$requires"
    test_in "$work_dir" "$const_pkg .." "$this_pkg" "$requires"
}

function cmd_with_rhino {
    local requires="$opt_requires,skill,js"

    pkg_bootstrap "$const_work_dir"
    ensure_pkgs_ready "$const_work_dir" "$requires,rhino"

    local rhino_dir="$const_work_dir/rhino-build/"
    local product="$rhino_dir/jsk/rhino.js"

    test -f "$product" &&                               \
        die 3 "Build product $product already present."

    test_in "$const_work_dir" "$const_pkg" 'rhino' "$requires"

    test -f "$product" ||                       \
        die 3 "Build product $product missing."

    # Continue with fresh Rhino.
    opt_skill="$(cd "$rhino_dir" && echo "$PWD/skill")"
    preloaded_pkgs='testing'
    main "$@"
}

function build_jenkins_trigger {
    if test -z "$opt_triggers"; then
        return
    fi

    tr -d '\n ' <<EOF
    <hudson.tasks.BuildTrigger>
      <childProjects>${opt_triggers:1}</childProjects>
      <threshold>
        <name>SUCCESS</name>
        <ordinal>0</ordinal>
        <color>BLUE</color>
      </threshold>
    </hudson.tasks.BuildTrigger>
EOF
}

function cmd_jenkins_job {
    parse_options 'jenkins-job' "$@"

    pkg_set_globals

    local job_in="$(dirname "$0")/misc/jenkins-job.xml.in"

    local pkg_repo="$opt_clone_base/$this_pkg.git"
    local testing_repo="$opt_clone_base/testing.git"
    local work_dir="$const_work_dir"
    local result_xml="$const_xml"
    local helper_opts=
    local trigger="$(build_jenkins_trigger)"

    for pkg in ${opt_requires//,/ }; do
        helper_opts="$helper_opts --require '$pkg'"
    done

    exec sed                                    \
        -e "s|@@PKG_REPO@@|$pkg_repo|"          \
        -e "s|@@TESTING_REPO@@|$testing_repo|"  \
        -e "s|@@WORK_DIR@@|$work_dir|"          \
        -e "s|@@HELPER_OPTS@@|$helper_opts|"    \
        -e "s|@@RESULT_XML@@|$result_xml|"      \
        -e "s|@@TRIGGER@@|$trigger|"            \
        < "$job_in"
}

function main {
    case "$1" in
        test)
            shift
            cmd_test "$@"
            ;;
        test-in)
            shift
            cmd_test_in "$@"
            ;;
        with-rhino)
            shift
            cmd_with_rhino "$@"
            ;;
        jenkins-job)
            shift
            cmd_jenkins_job "$@"
            ;;
        *)
            die_usage 2 "Unrecognized command: $*"
            ;;
    esac
}

main "$@"
